package net.sf.sido.support;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import net.sf.sido.DOInvalidPropertyTypeException;
import net.sf.sido.DOInvalidPropertyValueException;
import net.sf.sido.DONullPropertyException;
import net.sf.sido.DataObject;
import net.sf.sido.MetaInfo;
import net.sf.sido.PropertyAdapter;
import net.sf.sido.PropertyController;
import net.sf.sido.PropertyMetaInfos;

import com.google.common.base.Predicate;
import com.google.common.collect.ImmutableList;

public final class DefaultProperty<T> extends AbstractProperty<T> {

	private final List<PropertyController<T>> controllers;
	private final PropertyAdapter<T> adapter;
	private final MetaInfo metaInfo;

	public DefaultProperty(String name, boolean nullable, Class<? super T> type, boolean extension, Predicate<T> constraint, Collection<? extends PropertyController<T>> controllers) {
		this(name, nullable, type, extension, constraint, controllers, null);
	}

	public DefaultProperty(String name, boolean nullable, Class<? super T> type, boolean extension, Predicate<T> constraint, Collection<? extends PropertyController<T>> controllers,
			PropertyAdapter<T> adapter) {
		super(name, nullable, type, extension, constraint);
		this.adapter = adapter;
		this.controllers = ImmutableList.copyOf(controllers);
		Map<String, Object> mi = new HashMap<String, Object>();
		for(PropertyController<T> controller : this.controllers) {
			controller.fillMetaInfo(mi);
		}
		this.metaInfo = new DefaultMetaInfo(mi);
	}

	public DefaultProperty(String name, boolean nullable, Class<? super T> type, boolean extension, Predicate<T> constraint, PropertyController<T> controller) {
		this(name, nullable, type, extension, constraint, Collections.singletonList(controller));
	}

	@Override
	public T apply(DataObject o, T value) {
		// Default controls
		// Nullable property
		if (value == null && !isNullable()) {
			throw new DONullPropertyException(o, getName());
		}
		// For not null values only
		if (value != null) {
			// Adapts the value
			value = adapt(value);
			// Checks the type
			Class<? super T> type = getType();
			if (!type.isAssignableFrom(value.getClass())) {
				throw new DOInvalidPropertyTypeException(o, getName(), type.getName(), value);
			}
			// Checks the constraint
			Predicate<T> constraint = getConstraint();
			if (constraint != null && !constraint.apply(value)) {
				throw new DOInvalidPropertyValueException(o, getName(), value);
			}
			// Additional controls
			for(PropertyController<T> controller : this.controllers) {
				controller.control(this, o, value);
			}
			// OK
			return value;
		} else {
			return null;
		}
	}

	protected T adapt(T value) {
		if (adapter != null && adapter.canConvert(value)) {
			return adapter.convert(value);
		} else {
			return value;
		}
	}

	@Override
	public T getDefaultValue(boolean deep) {
		T value = null;
		Iterator<PropertyController<T>> i = controllers.iterator();
		while (value == null && i.hasNext()) {
			PropertyController<T> controller = i.next();
			value = controller.getDefaultValue(deep);
		}
		return value;
	}

	@Override
	public MetaInfo getMetaInfo() {
		return this.metaInfo;
	}

	@Override
	public boolean isArray() {
		return this.metaInfo.getBoolean(PropertyMetaInfos.ARRAY, false);
	}

	@Override
	public boolean isDO() {
		return metaInfo.getBoolean(PropertyMetaInfos.DO, false);
	}

	@Override
	public boolean isSimple() {
		return !isDO();
	}

}
